// Adapted from https://sourceforge.net/p/dadisi/code/8/tree/trunk/dadisi/
/*
 * File:   DistRandSlice.h
 * Author: amiranda
 *
 * Created on 8, November 2010, 12:24
 */

#ifndef _DISTRANDSLICE_H
#define _DISTRANDSLICE_H


#include <unordered_map>
#include <vector>
#define __STDC_LIMIT_MACROS // required for UINT64_MAX macro
#include <stdint.h>
#include <global/random_slicing/Disk.hpp>
#include <flat_segment_tree/flat_segment_tree.hpp>

// forward declaration for friend class
namespace gkfs::rpc {
class RandomSlicingDistributor;
}

#define DEBUG
#ifdef DEBUG
// #   define DUMP_COPIES
// #   define DUMP_DISKS
// #   define DUMP_PARTITIONS
// #   define DUMP_FREE_SPACE_COLLECTION
// #   define DUMP_FREE_SPACE_ASSIMILATION
// #   define DUMP_FREE_SPACE
// #   define DUMP_INTERVALS
// #   define DUMP_INTERVALS_VERBOSE
#endif


namespace VDRIVE {

/**
 * XXX: Place comment describing the distribution algorithm here
 *
 *
 *
 *
 */
class DistRandSlice {
public:
    static bool constexpr DEBUG_VSPACE = false;
    // constants declaring the min and max values for the
    // virtual allocation space.
    // right now, it spans from 0 to 2^64-1 blocks/files/whatever
    static uint64_t constexpr VSPACE_MIN = 0;
    static uint64_t constexpr VSPACE_MAX =
            DEBUG_VSPACE ? 1000000
                         : UINT64_MAX; // HARD DEBUG TODO(dauer) UINT64_MAX;

    static double constexpr DOUBLE_ALLOWED_ERROR =
            DEBUG_VSPACE ? 1.0e-5 : 1.0e-14;

    /**
     * Allow direct access e.g. to m_disks to avoid copying and memory
     * management issues.
     */
    friend class gkfs::rpc::RandomSlicingDistributor;

    /**
     * generate a new, uninitialized Rand Slice Implementation with default
     * settings.
     */
    DistRandSlice();

    /**
     * generate a new, uninitialized Rand Slice Implementation.
     */
    DistRandSlice(bool use_even_odd_collection, bool use_sorted_assimilation);

    /**
     * copy constructor
     *
     * @param orig original DistRandSlice
     */
    DistRandSlice(const DistRandSlice& orig);

    /**
     * Destructor
     */
    virtual ~DistRandSlice();

    /**
     * @see Distributor::placeExtent
     */
    virtual std::list<Disk*>*
    placeExtent(int64_t position);

    /**
     * @see Distributor::setConfiguration
     */
    virtual void
    setConfiguration(std::list<Disk*>* disks, int64_t extentsize,
                     int32_t copies);

    /**
     * @see Distributor::addDisks
     */
    virtual void
    addDisks(std::list<Disk*>* disks);

    /**
     * Added for GekkoFS - Remove the Disks given by their IDs
     * from the distribution.
     * TODO(dauer): Decide & document freeing of memory
     */
    void
    removeDisks(std::vector<int64_t> disk_ids);

    /**
     * @see Distributor::getDisks
     */
    virtual std::list<Disk*>*
    getDisks() const;

    /**
     * @see Distributor::getExtentsize
     */
    virtual int64_t
    getExtentsize() const {
        return m_extentsize;
    }

    /**
     * @see Distributor::getCopies
     */
    virtual int32_t
    getCopies() const {
        return m_copies;
    }

    /**
     * @return the number of partitions currently in the system
     */
    uint64_t
    getNumPartitions() const {
        return m_num_partitions;
    }

    /**
     * @return the number of intervals currently in the system
     */
    uint64_t
    getNumIntervals() const;

    /**
     * Deserialize from file.
     */
    void
    from_file(std::string filename);

    /**
     * Serialize to file.
     */
    void
    to_file(std::string filename);

private:
    void
    cleanup(void);

    void
    create_partitions(std::list<Disk*>* disks);

    void
    add_partitions(std::list<Disk*>* disks);

    uint64_t
    partition_capacity(uint64_t part_id, uint64_t new_total_capacity);

    uint64_t
    partition_capacity(Disk* disk, uint64_t new_total_capacity);

    void
    redistribute(
            std::unordered_map<uint64_t, uint64_t>& old_partitions,
            const std::list<std::pair<uint64_t, uint64_t>>& new_partitions);

    void
    collect_free_space(std::unordered_map<uint64_t, uint64_t>& old_partitions,
                       std::list<std::pair<uint64_t, uint64_t>>& free_space);

    void
    collect_free_space_even_odd(
            std::unordered_map<uint64_t, uint64_t>& old_partitions,
            std::list<std::pair<uint64_t, uint64_t>>& free_space);

    void
    reuse_free_space(
            std::list<std::pair<uint64_t, uint64_t>>& free_space,
            const std::list<std::pair<uint64_t, uint64_t>>& new_partitions);

    void
    reuse_free_space_sort(
            std::list<std::pair<uint64_t, uint64_t>>& free_space,
            const std::list<std::pair<uint64_t, uint64_t>>& new_partitions);


#if defined DEBUG
    void
    dump_intervals(void);
    void
    dump_free_space(const std::list<std::pair<uint64_t, uint64_t>>& l) const;
    void
    verify_partitions(void);
    std::unordered_map<uint64_t, uint64_t>
    compute_interval_sizes(void);
#endif

    /**
     * ExtenSize as given to setConfiguration.
     */
    int64_t m_extentsize;

    /**
     * number of copies to be distributed.
     */
    int32_t m_copies;

    /**
     * the disks as given to addDisks (or setConfiguration)
     */
    std::unordered_map<uint64_t, Disk*>* m_disks;

    /**
     * number of Disks contained by this Distributor.
     */
    uint64_t m_num_disks;

    /**
     * partition info: id -> capacity (Scaled relative to VSPACE_MAX. This
     * unit is also referred to as blocks.)
     */
    std::unordered_map<uint64_t, uint64_t>* m_partitions;

    /**
     * number of partitions
     */
    uint64_t m_num_partitions;

    /**
     * interval tree for searches
     */
    typedef ::mdds::flat_segment_tree<uint64_t, uint64_t> flat_segment_tree;
    flat_segment_tree* m_interval_tree;

    /**
     * capacity of the system (absolute i.e. sum of disk capacities)
     */
    uint64_t m_capacity;


    /**
     * use Even-Odd strategy when collecting free space
     */
    bool m_use_even_odd_collection;

    /**
     * sort free intervals decreasingly when assimilating free space
     */
    bool m_use_sorted_assimilation;


    /**
     * Init value for tree creation, needed for deserialization only.
     */
    uint64_t m_tree_init_value;
};
} // namespace VDRIVE
#endif /* _DISTRANDSLICE_H */
