/*
  Copyright 2018-2020, Barcelona Supercomputing Center (BSC), Spain
  Copyright 2015-2020, Johannes Gutenberg Universitaet Mainz, Germany

  This software was partially supported by the
  EC H2020 funded project NEXTGenIO (Project ID: 671951, www.nextgenio.eu).

  This software was partially supported by the
  ADA-FS project under the SPPEXA project funded by the DFG.

  SPDX-License-Identifier: MIT
*/


#include <daemon/relocation/config_manager.hpp>
#include <daemon/daemon.hpp>
#include <global/global_defs.hpp>
#include <daemon/handler/rpc_defs.hpp>
#include <daemon/handler/rpc_util.hpp>
#include <global/hostsfile_util.hpp>
#include <global/hostsfile_util.hpp>
#include <global/random_slicing/DistRandSlice.hpp>
#include <global/rpc/rpc_types.hpp>
#include <global/rpc/distributor.hpp>

#include <cassert>
#include <filesystem>
#include <iostream>
#include <list>

#include <fmt/format.h>

// margo
extern "C" {
#include <margo.h>
}

using namespace std;

namespace gkfs::relocation {

void
calculate_random_slicing() {
    using gkfs::rpc::host_t;

    auto hosts = read_hosts_file();
    list<VDRIVE::Disk*> disks;
    for(const auto& [id, host] : hosts) {
        const auto& [hostname, uri] = host;
        disks.emplace_back(new VDRIVE::Disk(id, 1, hostname, uri));
        cout << uri << "\n";
    }
    auto filename = GKFS_DATA->hosts_file() + ".rsconfig";
    auto dist = std::make_shared<VDRIVE::DistRandSlice>();
    if(std::filesystem::exists(filename)) {
        GKFS_DATA->spdlogger()->info(
                "Existing random slicing configuration found, "
                "loading from file: {}",
                filename);
        dist->from_file(filename);

        // update configuration

        // create distributor
        gkfs::rpc::RandomSlicingDistributor rsdist(dist);

        // get vector of all existing rs hosts
        auto all_rs_hosts = rsdist.locate_directory_metadata("");
        vector<host_t> all_hostsfile_hosts{};
        for(const auto& host : hosts) {
            all_hostsfile_hosts.push_back(host.first);
        }

        // compare with read hosts
        vector<host_t> gone_hosts{};
        vector<host_t> new_hosts{};

        // lists need to be sorted for set_difference
        std::sort(all_rs_hosts.begin(), all_rs_hosts.end());
        std::sort(all_hostsfile_hosts.begin(), all_hostsfile_hosts.end());

        // gone = rs_hosts \ hostsfile_hosts
        std::set_difference(all_rs_hosts.begin(), all_rs_hosts.end(),
                            all_hostsfile_hosts.begin(),
                            all_hostsfile_hosts.end(),
                            std::inserter(gone_hosts, gone_hosts.begin()));

        // new = hostsfile_hosts \ rs_hosts
        std::set_difference(all_hostsfile_hosts.begin(),
                            all_hostsfile_hosts.end(), all_rs_hosts.begin(),
                            all_rs_hosts.end(),
                            std::inserter(new_hosts, new_hosts.begin()));
        // remove gone hosts
        // add new hosts
        // write to file
        // be done with it
    } else {
        GKFS_DATA->spdlogger()->info(
                "No existing random slicing configuration found, "
                "creating from scratch with {} hosts.",
                hosts.size());
        dist->setConfiguration(&disks, 1, 1);
    }
    for(auto disk : disks) {
        delete disk;
    }

    dist->to_file(filename);

    // Just test reading here
    VDRIVE::DistRandSlice dist_read{};
    dist_read.from_file(filename);
    dist_read.to_file(filename + ".reexport");
}

unordered_map<uint64_t, pair<string, string>>
read_hosts_file() {
    string hostfile = GKFS_DATA->hosts_file();

    unordered_map<uint64_t, pair<string, string>> hosts;
    try {
        hosts = gkfs::utils::load_hostfile(hostfile);
    } catch(const exception& e) {
        auto emsg = fmt::format("Failed to load hosts file: {}", e.what());
        throw runtime_error(emsg);
    }

    if(hosts.empty()) {
        throw runtime_error(fmt::format("Hostfile empty: '{}'", hostfile));
    }

    GKFS_DATA->spdlogger()->info("Hosts pool size: {}", hosts.size());

    return hosts;
}

/**
 * Send RPCs to all daemons in hosts file to trigger relocation.
 */
void
do_relocation() {
    auto hosts = read_hosts_file();

    cout << "number of hosts: " << hosts.size() << "\n";
    auto mid =
            margo_init(GKFS_DATA->bind_addr().c_str(), MARGO_CLIENT_MODE, 0, 0);
    if(mid == MARGO_INSTANCE_NULL) {
        throw runtime_error("Failed to initialize the Margo RPC client");
    }

    hg_id_t relocation_start_id = MARGO_REGISTER(
            mid, gkfs::rpc::tag::relocation_start, rpc_relocation_start_in_t,
            rpc_err_out_t, rpc_srv_relocation_start);


    for(const auto& [host_id, host] : hosts) {
        const auto& [hostname, uri] = host;
        cout << fmt::format("Hostname: {} Address: {} ID: {}\n", hostname, uri,
                            host_id);
        hg_addr_t host_addr;
        rpc_relocation_start_in_t in{};
        rpc_err_out_t out{};
        in.host_id = host_id;
        auto ret = margo_addr_lookup(mid, uri.c_str(), &host_addr);
        assert(ret == HG_SUCCESS);

        // let's do this sequential first
        hg_handle_t handle;
        ret = margo_create(mid, host_addr, relocation_start_id, &handle);
        assert(ret == HG_SUCCESS);

        ret = margo_forward(handle, &in); // blocking
        assert(ret == HG_SUCCESS);

        ret = margo_get_output(handle, &out);
        assert(ret == HG_SUCCESS);

        // TODO(dauer) process output

        if(HG_SUCCESS !=
           gkfs::rpc::margo_client_cleanup(&handle, &out, &mid, &host_addr)) {
            cout << "Error during margo cleanup.\n";
        }
    }
    margo_finalize(mid);
}

} // namespace gkfs::relocation