/*
  Copyright 2018-2020, Barcelona Supercomputing Center (BSC), Spain
  Copyright 2015-2020, Johannes Gutenberg Universitaet Mainz, Germany

  This software was partially supported by the
  EC H2020 funded project NEXTGenIO (Project ID: 671951, www.nextgenio.eu).

  This software was partially supported by the
  ADA-FS project under the SPPEXA project funded by the DFG.

  SPDX-License-Identifier: MIT
*/

#include <daemon/relocation/transmitter.hpp>

#include <daemon/backend/data/chunk_storage.hpp>
#include <daemon/backend/metadata/db.hpp>
#include <daemon/daemon.hpp>
#include <daemon/handler/rpc_defs.hpp>
#include <daemon/handler/rpc_util.hpp>
#include <global/hostsfile_util.hpp>
#include <daemon/relocation/config_manager.hpp>
#include <global/rpc/rpc_types.hpp>

#include <filesystem>
namespace fs = std::filesystem;
#include <iostream>

#include <fmt/format.h>

using namespace std;

/*
 * This file contains functions to send metadata and data to other daemons
 * during relocation. This daemon acts as a Margo RPC client connecting to
 * those daemons.
 */

namespace gkfs::relocation {


void
transmit_metadata_and_data(gkfs::rpc::host_t localhost) {
    {
        // TODO(dauer) move this to somewhere appropriate
        auto dist_impl = make_shared<VDRIVE::DistRandSlice>();
        dist_impl->from_file(GKFS_DATA->hosts_file() + ".rsconfig");
        auto distributor =
                make_shared<gkfs::rpc::RandomSlicingDistributor>(dist_impl);
        GKFS_DATA->distributor(distributor);
    }

    auto hosts = read_hosts_file();
    auto mid = RPC_DATA->client_rpc_mid();
    auto relocate_metadata_id =
            gkfs::rpc::get_rpc_id(mid, gkfs::rpc::tag::relocate_metadata);

    auto& distributor = *(GKFS_DATA->distributor());
    cout << fmt::format("Got host_id = {} and parsed {} hosts.\n", localhost,
                        hosts.size());

    // Relocate metadata
    for(auto metadentry : GKFS_DATA->mdb()->get_all()) {
        if(metadentry.first == "/") {
            continue;
        }
        auto destination = distributor.locate_file_metadata(metadentry.first);
        cout << "Metadentry " << metadentry.first << " : " << metadentry.second
             << (destination == localhost
                         ? " Stay"
                         : " -> Goto "s + std::to_string(destination))
             << '\n';

        if(destination == localhost) {
            continue;
        }
        // send RPC
        rpc_relocate_metadata_in_t in{};
        rpc_err_out_t out{};
        hg_addr_t host_addr{};

        in.key = metadentry.first.c_str();
        in.value = metadentry.second.c_str();

        auto ret = margo_addr_lookup(mid, hosts[destination].second.c_str(),
                                     &host_addr);
        assert(ret == HG_SUCCESS);

        // let's do this sequential first
        hg_handle_t handle;
        ret = margo_create(mid, host_addr, relocate_metadata_id, &handle);
        assert(ret == HG_SUCCESS);

        ret = margo_forward(handle, &in); // blocking
        assert(ret == HG_SUCCESS);

        ret = margo_get_output(handle, &out);
        assert(ret == HG_SUCCESS);

        // TODO(dauer) process output
        GKFS_DATA->mdb()->remove(in.key);

        if(HG_SUCCESS !=
           gkfs::rpc::margo_client_cleanup(&handle, &out, &mid, &host_addr)) {
            cout << "Error during margo cleanup.\n";
        }
    }

    // Relocate data (chunks)
    auto relocate_chunk_id =
            gkfs::rpc::get_rpc_id(mid, gkfs::rpc::tag::relocate_chunk);
    for(auto& chunks_dir : GKFS_DATA->storage()->chunks_directory_iterator()) {
        assert(chunks_dir.is_directory()); // TODO(dauer) just log
        string file_path = GKFS_DATA->storage()->get_file_path(
                chunks_dir.path().filename().string());

        for(auto& chunk_file : fs::directory_iterator(chunks_dir)) {
            assert(chunk_file.is_regular_file()); // TODO(dauer) just log
            gkfs::rpc::chnk_id_t chunk_id =
                    std::stoul(chunk_file.path().filename().string());
            auto destination = distributor.locate_data(file_path, chunk_id);
            size_t size = chunk_file.file_size();
            cout << file_path << " chunk: " << chunk_id << " size: " << size
                 << (destination == localhost
                             ? " Stay"
                             : " -> Goto "s + std::to_string(destination))
                 << '\n';

            if(destination == localhost) {
                continue;
            }

            // prepare bulk
            unique_ptr<char[]> buf(new char[size]());
            // read data (blocking)
            hg_size_t bytes_read = GKFS_DATA->storage()->read_chunk(
                    file_path, chunk_id, buf.get(), size, 0);
            hg_bulk_t bulk{};
            char* bufptr = buf.get();
            auto ret = margo_bulk_create(mid, 1, (void**) &bufptr, &bytes_read,
                                         HG_BULK_READ_ONLY, &bulk);
            assert(ret == HG_SUCCESS);

            // send RPC
            rpc_relocate_chunk_in_t in{};
            rpc_err_out_t out{};
            hg_addr_t host_addr{};

            in.path = file_path.c_str();
            in.chunk_id = chunk_id;
            in.bulk_handle = bulk;

            ret = margo_addr_lookup(mid, hosts[destination].second.c_str(),
                                    &host_addr);
            assert(ret == HG_SUCCESS);

            // let's do this sequential first
            hg_handle_t handle;
            ret = margo_create(mid, host_addr, relocate_chunk_id, &handle);
            assert(ret == HG_SUCCESS);

            ret = margo_forward(handle, &in); // blocking
            assert(ret == HG_SUCCESS);

            ret = margo_get_output(handle, &out);
            assert(ret == HG_SUCCESS);

            // TODO(dauer) process output
            GKFS_DATA->storage()->remove_chunk(file_path, chunk_id);

            if(HG_SUCCESS != gkfs::rpc::margo_client_cleanup(
                                     &handle, &out, &mid, &host_addr, &bulk)) {
                cout << "Error during margo cleanup.\n";
            }
        }
    }
}


} // namespace gkfs::relocation