/*
  Copyright 2018-2020, Barcelona Supercomputing Center (BSC), Spain
  Copyright 2015-2020, Johannes Gutenberg Universitaet Mainz, Germany

  This software was partially supported by the
  EC H2020 funded project NEXTGenIO (Project ID: 671951, www.nextgenio.eu).

  This software was partially supported by the
  ADA-FS project under the SPPEXA project funded by the DFG.

  SPDX-License-Identifier: MIT
*/

#include <global/rpc/distributor.hpp>

using namespace std;

namespace gkfs::rpc {


// SimpleHashDistributor

SimpleHashDistributor::SimpleHashDistributor(host_t localhost,
                                             unsigned int hosts_size)
    : localhost_(localhost), hosts_size_(hosts_size), all_hosts_(hosts_size) {
    ::iota(all_hosts_.begin(), all_hosts_.end(), 0);
}

host_t
SimpleHashDistributor::localhost() const {
    return localhost_;
}

unsigned int
SimpleHashDistributor::hosts_size() const {
    return hosts_size_;
}

host_t
SimpleHashDistributor::locate_data(const string& path,
                                   const chunkid_t& chnk_id) const {
    return str_hash(path + ::to_string(chnk_id)) % hosts_size_;
}

host_t
SimpleHashDistributor::locate_file_metadata(const string& path) const {
    return str_hash(path) % hosts_size_;
}

::vector<host_t>
SimpleHashDistributor::locate_directory_metadata(const string& path) const {
    return all_hosts_;
}


// LocalOnlyDistributor (seems to be unused - TODO remove?)

LocalOnlyDistributor::LocalOnlyDistributor(host_t localhost)
    : localhost_(localhost) {}

host_t
LocalOnlyDistributor::localhost() const {
    return localhost_;
}

host_t
LocalOnlyDistributor::locate_data(const string& path,
                                  const chunkid_t& chnk_id) const {
    return localhost_;
}

host_t
LocalOnlyDistributor::locate_file_metadata(const string& path) const {
    return localhost_;
}

::vector<host_t>
LocalOnlyDistributor::locate_directory_metadata(const string& path) const {
    return {localhost_};
}


// ForwarderDistributor

ForwarderDistributor::ForwarderDistributor(host_t fwhost,
                                           unsigned int hosts_size)
    : fwd_host_(fwhost), hosts_size_(hosts_size), all_hosts_(hosts_size) {
    ::iota(all_hosts_.begin(), all_hosts_.end(), 0);
}

host_t
ForwarderDistributor::localhost() const {
    return fwd_host_;
}

unsigned int
ForwarderDistributor::hosts_size() const {
    return hosts_size_;
}

host_t
ForwarderDistributor::locate_data(const std::string& path,
                                  const chunkid_t& chnk_id) const {
    return fwd_host_;
}

host_t
ForwarderDistributor::locate_file_metadata(const std::string& path) const {
    return str_hash(path) % hosts_size_;
}

std::vector<host_t>
ForwarderDistributor::locate_directory_metadata(const std::string& path) const {
    return all_hosts_;
}


// RandomSlicingDistributor for dynamic placement

RandomSlicingDistributor::RandomSlicingDistributor(
        host_t localhost, std::shared_ptr<VDRIVE::DistRandSlice> dist_impl)
    : localhost_(localhost), all_hosts_(0), dist_impl_(dist_impl) {
    for(const auto& disk : *dist_impl_->m_disks) {
        all_hosts_.push_back(disk.second->getId());
    }
}

RandomSlicingDistributor::RandomSlicingDistributor(
        std::shared_ptr<VDRIVE::DistRandSlice> dist_impl)
    : RandomSlicingDistributor(-1, dist_impl) {}

host_t
RandomSlicingDistributor::localhost() const {
    assert(false); // localhost is probably not set. Who calls this anyway on
                   // this distributor?
    return localhost_;
}

unsigned int
RandomSlicingDistributor::hosts_size() const {
    return all_hosts_.size();
}

host_t
RandomSlicingDistributor::locate_data(const std::string& path,
                                      const chunkid_t& chnk_id) const {
    return locate_file_metadata(path + ::to_string(chnk_id));
}

host_t
RandomSlicingDistributor::locate_file_metadata(const std::string& path) const {
    auto myhash = str_hash(path);
    // TODO placeExtent hashes the hash with it's own hash, is this a problem?
    // (migh be unavoidable if multiple copies should be stored sometime)
    auto disklist = dist_impl_->placeExtent(myhash);
    assert(disklist->size() == 1);
    auto disk = disklist->front();
    auto id = disk->getId();
    delete disk;
    delete disklist;
    // TODO Avoid memory location for this task and make this much simpler
    return id;
}

std::vector<host_t>
RandomSlicingDistributor::locate_directory_metadata(
        const std::string& path) const {
    return all_hosts_;
}

std::unordered_map<uint64_t, std::pair<std::string, std::string>>
RandomSlicingDistributor::get_hosts_map() {
    std::unordered_map<uint64_t, std::pair<std::string, std::string>> hosts{};
    for(const auto& pair : *(dist_impl_->m_disks)) {
        VDRIVE::Disk* disk = pair.second;
        hosts[disk->getId()] = make_pair(disk->getHostname(), disk->getUri());
    }
    return hosts;
}
} // namespace gkfs::rpc
